#include <string>
#include <list>
#include <map>
#include <set>
#include <cmath>
#include <fstream>

#define IFTY 100000

// Forward declaration of vertex class
class vertex;

class edge
{
public:
	vertex *u, *v;
	int weight;

	edge(vertex* a,vertex* b,int w)
	{
		u = a; v = b; weight = w;
	}
};

// Each vertex contains within it a list of edges
// to the other vertices the vertex is connected to.
class vertex
{
public:
	int key;
	double lon, lat;
	double d;
	vertex *p;
	std::list<edge> edges;

	vertex() {}
	vertex(int v) { key = v; d = IFTY; p = NULL; }
};

class graph
{
public:
	graph();
	~graph();

	void dijkstra(int s);
	
	void pathTo(int dest,std::list<int>& L);

    int getVertexCount() { return V.size(); }
	vertex* getVertex(int n) { return V[n]; }
private:
	// Since the edges are stored in the vertices, we only need
	// a list of vertices in the graph. For convenience, this is
	// stored as a map.
	std::map<int,vertex*> V;
};

graph::graph()
{
	std::ifstream in;
	in.open("graph.txt");
	
	int nV;
	in >> nV;
	int key;
		
	for(int n = 0;n < nV;n++) {
		in >> key;
		V[key] = new vertex(key);
	}
	
	int nE;
	in >> nE;
	int from, to, w;
	for(int n = 0;n < nE;n++) {
		in >> from >> to >> w;
		vertex *u = V[from];
		vertex *v = V[to];
		u->edges.push_back(edge(u, v, w));
		v->edges.push_back(edge(v, u, w));
	}
	in.close();
}

graph::~graph() {
	auto itr = V.begin();
	while(itr != V.end()) {
		delete itr->second;
		itr++;
	}
}

// Helper class for Dijkstra's algorithm
// We need this because we are going to be storing
// pointers to vertices in a set in dijkstra. The
// set will need a comparator object to determine
// the order of elements in the set: we will want to
// order these pointers by the d values of the vertices
// that the pointers point to.
// A further hack here is some logic that treats
// two vertices with the same d value as different. We
// need this to work around a key characteristic of the
// STL set class, which is that STL set does not allow
// us to insert duplicates.
class vertex_less
{
public:
	bool operator()(vertex* a,vertex* b) const
	{ if(a->d == b->d) return a < b; else return a->d < b->d; }
};

void graph::dijkstra(int s) {

	// Dijkstra's algorithm is driven by a priority queue.
	// Since the STL priority queue does not offer support
	// for the decrease key operation, we implement the 
	// priority queue via a set. Since the set class orders
	// its contents, we can always find the minimum priority
	// element in the set by calling begin().
	std::set<vertex*,vertex_less > Q;
	
	auto itr = V.begin();
	while(itr != V.end()) {
		if(itr->first != s) {
			itr->second->d = INFINITY;
			itr->second->p = NULL;
		} else {
			itr->second->d = 0;
			itr->second->p = itr->second;
		}
		Q.insert(itr->second);
		itr++;
	}

	while(!Q.empty()) {
		vertex *u = *(Q.begin());
		Q.erase(Q.begin());

		std::list<edge>& edges = u->edges;
		auto itr2 = edges.begin();
		while(itr2 != edges.end()) {
			vertex* v = itr2->v;
			if(u->d + itr2->weight < v->d) {
				// Since the STL set class does not support a
				// decrease key operation, we simulate a decrease
				// key by removing v, decreasing its key, and then
				// reinserting it in the set. Since erase, find,
				// and insert are all O(lg(N)) operations in a set,
				// this is almost as fast as a legitimate decrease
				// key operation in a binomial heap.
				Q.erase(Q.find(v));
				v->d = u->d + itr2->weight;
				v->p = u;
				Q.insert(v);
			}
			itr2++;
		}
	}
}

void graph::pathTo(int dest,std::list<int>& L)
{
	vertex *v = V[dest];
	if(v->d == 0)
		L.push_back(dest);
	else if(v->p == NULL)
		return;
	else {
		this->pathTo(v->p->key,L);
		L.push_back(dest);
	}
}
